/* File: ast.cc
 * ------------
 */

#include "ast.h"
#include "ast_type.h"
#include "ast_decl.h"
#include "symtable.h"
#include <string.h> // strdup
#include <stdio.h>  // printf


Node::Node(yyltype loc) {
    location = new yyltype(loc);
    parent = NULL;
}

Node::Node() {
    location = NULL;
    parent = NULL;
}

/* The Print method is used to print the parse tree nodes.
 * If this node has a location (most nodes do, but some do not), it
 * will first print the line number to help you match the parse tree 
 * back to the source text. It then indents the proper number of levels 
 * and prints the "print name" of the node. It then will invoke the
 * virtual function PrintChildren which is expected to print the
 * internals of the node (itself & children) as appropriate.
 */
void Node::Print(int indentLevel, const char *label) { 
    const int numSpaces = 3;
    printf("\n");
    if (GetLocation()) 
        printf("%*d", numSpaces, GetLocation()->first_line);
    else 
        printf("%*s", numSpaces, "");
    printf("%*s%s%s: ", indentLevel*numSpaces, "", 
           label? label : "", GetPrintNameForNode());
   PrintChildren(indentLevel);
} 
	 
Identifier::Identifier(yyltype loc, const char *n) : Node(loc) {
    name = strdup(n);
} 

void Identifier::PrintChildren(int indentLevel) {
    printf("%s", name);
}

/*------------------------------------------------------------------------------------
 * Symbol Table Information
 *----------------------------------------------------------------------------------*/
SymbolTable symbols;
pair<Decl*,llvm::Value *> findSymbol(string s) {
    for(int i = symbols.size() - 1; i >= 0; i--) {
        if(symbols[i].count(s) > 0) return symbols[i][s];
    }
    return make_pair<Decl*,llvm::Value *>(NULL,NULL);
}

pair<Decl*,llvm::Value *> findSymbol(Decl* d) {
    string s(d->GetIdentifier()->GetName());
    return findSymbol(s);
}

bool isInCurrentScope(string s) {
    if(symbols.size() == 0) return false;
    return symbols[symbols.size()-1].count(s);
}

bool isInCurrentScope(Decl* d) {
    string s(d->GetIdentifier()->GetName());
    return isInCurrentScope(s);
}

bool isInScope(string s) {
    pair<Decl*, llvm::Value *> tmp = findSymbol(s);
    return tmp.first != NULL;
}

bool isInScope(Decl* d) {
    string s(d->GetIdentifier()->GetName());
    return isInScope(s);
}

void pushScope() {
    symbols.push_back(map<string,pair<Decl*,llvm::Value *> >());
}
void popScope() {
    symbols.pop_back();
}

void addToScope(string s, Decl* d, llvm::Value * v) {
    symbols[symbols.size()-1][s] = make_pair(d, v);
}

bool global(){
    return symbols.front()==symbols.back();
}
