/* File: ast_decl.cc
 * -----------------
 * Implementation of Decl node classes.
 */
#include "ast_decl.h"
#include "ast_type.h"
#include "ast_stmt.h"
#include "symtable.h"        
         
Decl::Decl(Identifier *n) : Node(*n->GetLocation()) {
    Assert(n != NULL);
    (id=n)->SetParent(this); 
}

VarDecl::VarDecl(Identifier *n, Type *t, Expr *e) : Decl(n) {
    Assert(n != NULL && t != NULL);
    (type=t)->SetParent(this);
    if (e) (assignTo=e)->SetParent(this);
    typeq = NULL;
}

VarDecl::VarDecl(Identifier *n, TypeQualifier *tq, Expr *e) : Decl(n) {
    Assert(n != NULL && tq != NULL);
    (typeq=tq)->SetParent(this);
    if (e) (assignTo=e)->SetParent(this);
    type = NULL;
}

VarDecl::VarDecl(Identifier *n, Type *t, TypeQualifier *tq, Expr *e) : Decl(n) {
    Assert(n != NULL && t != NULL && tq != NULL);
    (type=t)->SetParent(this);
    (typeq=tq)->SetParent(this);
    if (e) (assignTo=e)->SetParent(this);
}
  
void VarDecl::PrintChildren(int indentLevel) { 
   if (typeq) typeq->Print(indentLevel+1);
   if (type) type->Print(indentLevel+1);
   if (id) id->Print(indentLevel+1);
   if (assignTo) assignTo->Print(indentLevel+1, "(initializer) ");
}

FnDecl::FnDecl(Identifier *n, Type *r, List<VarDecl*> *d) : Decl(n) {
    Assert(n != NULL && r!= NULL && d != NULL);
    (returnType=r)->SetParent(this);
    (formals=d)->SetParentAll(this);
    body = NULL;
    returnTypeq = NULL;
}

FnDecl::FnDecl(Identifier *n, Type *r, TypeQualifier *rq, List<VarDecl*> *d) : Decl(n) {
    Assert(n != NULL && r != NULL && rq != NULL&& d != NULL);
    (returnType=r)->SetParent(this);
    (returnTypeq=rq)->SetParent(this);
    (formals=d)->SetParentAll(this);
    body = NULL;
}

void FnDecl::SetFunctionBody(Stmt *b) { 
    (body=b)->SetParent(this);
}

void FnDecl::PrintChildren(int indentLevel) {
    if (returnType) returnType->Print(indentLevel+1, "(return type) ");
    if (id) id->Print(indentLevel+1);
    if (formals) formals->PrintAll(indentLevel+1, "(formals) ");
    if (body) body->Print(indentLevel+1, "(body) ");
}

llvm::Value* VarDecl::Emit() {
    llvm::BasicBlock* b = irgen.GetBasicBlock();
    llvm::Value* val;
    //locals
    if(b) {
        if(dynamic_cast<ArrayType *>(type) !=NULL){
            llvm::Value* val2 = llvm::ConstantInt::get(irgen.GetIntType(), dynamic_cast<ArrayType*>(type)->GetElemCount());
            llvm::Type* thisType = convertType(dynamic_cast<ArrayType*>(type)->GetElemType());
            llvm::ArrayType* arrType = llvm::ArrayType::get(thisType, dynamic_cast<ArrayType*>(type)->GetElemCount());
            val = new llvm::AllocaInst(arrType, val2, id->GetName(), irgen.GetBasicBlock());
        }
        else{
            val = new llvm::AllocaInst(convertType(type), id->GetName(), b);
        }
        if(assignTo) new llvm::StoreInst(assignTo->Emit(), val, false, irgen.GetBasicBlock());
    }
    else {
        llvm::Constant* e = assignTo ? (llvm::Constant*)assignTo->Emit() : llvm::Constant::getNullValue(convertType(type));
        val = new llvm::GlobalVariable(*irgen.GetOrCreateModule("program"), convertType(type), false, llvm::GlobalValue::ExternalLinkage, e, llvm::Twine(id->GetName()));
    }
    addToScope(id->GetName(), this, val);
    return val;
}


llvm::Value * FnDecl::Emit(){
    llvm::Module *mod = irgen.GetOrCreateModule("program");
    llvm::Type *rt = convertType(this->returnType);

    std::vector<llvm::Type *> ref;
    for (int i = 0; i < formals->NumElements(); i++){
        ref.push_back(convertType(formals->Nth(i)->GetType()));
    }
    llvm::ArrayRef<llvm::Type *> ar(ref);
    llvm::FunctionType * ft = llvm::FunctionType::get(rt, ar, false);

    llvm::Value * funcval = mod->getOrInsertFunction(this->GetIdentifier()->GetName(), ft);
    llvm::Function * fp = llvm::cast<llvm::Function>(funcval);

    addToScope(id->GetName(), this, funcval);
    irgen.SetFunction(fp);

    //SetName
    pushScope();
    llvm::Function::arg_iterator args = fp->arg_begin();

    llvm::LLVMContext *context = irgen.GetContext();
    llvm::BasicBlock * varbb = llvm::BasicBlock::Create(*context, "entry", fp);
    irgen.SetBasicBlock(varbb);
    for (int i = 0; i < formals->NumElements(); args++, i++){
        VarDecl * currvd = formals->Nth(i);
        string currname = string(currvd->GetIdentifier()->GetName());
        args->setName(currname);
        llvm::Value * currargv = currvd->Emit();
        llvm::StoreInst * si = new llvm::StoreInst(&*args, currargv, varbb);
    }
    //Emit StmtBlock
    body->Emit();
    // we probably need something like this in other places but whatever for now
    llvm::BasicBlock* block = irgen.GetBasicBlock();
    if(!block->getTerminator()) {
        if(this->returnType != Type::voidType) {
            llvm::ReturnInst::Create(*context, llvm::Constant::getNullValue(convertType(this->returnType)), block);
        }
        else llvm::ReturnInst::Create(*context, block);
    }
    popScope();
    irgen.SetFunction(NULL);
    irgen.SetBasicBlock(NULL);
    return funcval;
}
