/* File: ast_expr.cc
 * -----------------
 * Implementation of expression node classes.
 */

#include <string.h>
#include "ast_expr.h"
#include "ast_type.h"
#include "ast_decl.h"
#include "symtable.h"


IntConstant::IntConstant(yyltype loc, int val) : Expr(loc) {
    value = val;
}

void IntConstant::PrintChildren(int indentLevel) {
    printf("%d", value);
}

FloatConstant::FloatConstant(yyltype loc, double val) : Expr(loc) {
    value = val;
}
void FloatConstant::PrintChildren(int indentLevel) {
    printf("%g", value);
}

BoolConstant::BoolConstant(yyltype loc, bool val) : Expr(loc) {
    value = val;
}
void BoolConstant::PrintChildren(int indentLevel) {
    printf("%s", value ? "true" : "false");
}

VarExpr::VarExpr(yyltype loc, Identifier *ident) : Expr(loc) {
    Assert(ident != NULL);
    this->id = ident;
}

void VarExpr::PrintChildren(int indentLevel) {
    id->Print(indentLevel+1);
}

Operator::Operator(yyltype loc, const char *tok) : Node(loc) {
    Assert(tok != NULL);
    strncpy(tokenString, tok, sizeof(tokenString));
}

void Operator::PrintChildren(int indentLevel) {
    printf("%s",tokenString);
}

bool Operator::IsOp(const char *op) const {
    return strcmp(tokenString, op) == 0;
}

CompoundExpr::CompoundExpr(Expr *l, Operator *o, Expr *r) 
  : Expr(Join(l->GetLocation(), r->GetLocation())) {
    Assert(l != NULL && o != NULL && r != NULL);
    (op=o)->SetParent(this);
    (left=l)->SetParent(this); 
    (right=r)->SetParent(this);
}

CompoundExpr::CompoundExpr(Operator *o, Expr *r) 
  : Expr(Join(o->GetLocation(), r->GetLocation())) {
    Assert(o != NULL && r != NULL);
    left = NULL; 
    (op=o)->SetParent(this);
    (right=r)->SetParent(this);
}

CompoundExpr::CompoundExpr(Expr *l, Operator *o) 
  : Expr(Join(l->GetLocation(), o->GetLocation())) {
    Assert(l != NULL && o != NULL);
    (left=l)->SetParent(this);
    (op=o)->SetParent(this);
}

void CompoundExpr::PrintChildren(int indentLevel) {
   if (left) left->Print(indentLevel+1);
   op->Print(indentLevel+1);
   if (right) right->Print(indentLevel+1);
}
   
ConditionalExpr::ConditionalExpr(Expr *c, Expr *t, Expr *f)
  : Expr(Join(c->GetLocation(), f->GetLocation())) {
    Assert(c != NULL && t != NULL && f != NULL);
    (cond=c)->SetParent(this);
    (trueExpr=t)->SetParent(this);
    (falseExpr=f)->SetParent(this);
}

void ConditionalExpr::PrintChildren(int indentLevel) {
    cond->Print(indentLevel+1, "(cond) ");
    trueExpr->Print(indentLevel+1, "(true) ");
    falseExpr->Print(indentLevel+1, "(false) ");
}
ArrayAccess::ArrayAccess(yyltype loc, Expr *b, Expr *s) : LValue(loc) {
    (base=b)->SetParent(this); 
    (subscript=s)->SetParent(this);
}

void ArrayAccess::PrintChildren(int indentLevel) {
    base->Print(indentLevel+1);
    subscript->Print(indentLevel+1, "(subscript) ");
}
     
FieldAccess::FieldAccess(Expr *b, Identifier *f)
  : LValue(b? Join(b->GetLocation(), f->GetLocation()) : *f->GetLocation()) {
    Assert(f != NULL); // b can be be NULL (just means no explicit base)
    base = b; 
    if (base) base->SetParent(this); 
    (field=f)->SetParent(this);
}


void FieldAccess::PrintChildren(int indentLevel) {
    if (base) base->Print(indentLevel+1);
    field->Print(indentLevel+1);
}

Call::Call(yyltype loc, Expr *b, Identifier *f, List<Expr*> *a) : Expr(loc)  {
    Assert(f != NULL && a != NULL); // b can be be NULL (just means no explicit base)
    base = b;
    if (base) base->SetParent(this);
    (field=f)->SetParent(this);
    (actuals=a)->SetParentAll(this);
}

void Call::PrintChildren(int indentLevel) {
   if (base) base->Print(indentLevel+1);
   if (field) field->Print(indentLevel+1);
   if (actuals) actuals->PrintAll(indentLevel+1, "(actuals) ");
}

llvm::Value* VarExpr::Emit() {
    return new llvm::LoadInst(findSymbol(id->GetName()).second, "", irgen.GetBasicBlock());
}

string Operator::getToken() const {
  return tokenString;
}

llvm::Value* IntConstant::Emit() {
    return llvm::ConstantInt::get(llvm::Type::getInt32Ty(*irgen.GetContext()), value, true);
}

llvm::Value* FloatConstant::Emit() {
    return llvm::ConstantFP::get(llvm::Type::getFloatTy(*irgen.GetContext()), value);
}

llvm::Value* BoolConstant::Emit() {
    return llvm::ConstantInt::get(llvm::Type::getInt1Ty(*irgen.GetContext()), value);
}

llvm::Value* Call::Emit() {
    vector<llvm::Value*> args;
    for(int i = 0; i < actuals->NumElements(); i++) {
        args.push_back(actuals->Nth(i)->Emit());
    }
    return llvm::CallInst::Create(findSymbol(field->GetName()).second, args, "", irgen.GetBasicBlock());
}

llvm::Value* ConditionalExpr::Emit() {
    llvm::LLVMContext * context = irgen.GetContext();
    llvm::Function * func = irgen.GetFunction();
    llvm::BasicBlock* b = irgen.GetBasicBlock();
    llvm::BasicBlock * thenBlock = llvm::BasicBlock::Create(*context, "thenBlock", func);
    llvm::BasicBlock * footBlock = llvm::BasicBlock::Create(*context, "footBlock", func);
    llvm::BasicBlock * elseBlock =  llvm::BasicBlock::Create(*context, "elseBlock", func);
    llvm::Value * condition = cond->Emit();

    irgen.SetBasicBlock(thenBlock);
    llvm::Value* tVal = trueExpr->Emit();
    llvm::Value* val = new llvm::AllocaInst(tVal->getType(), "", b);
    new llvm::StoreInst(tVal, val, false, irgen.GetBasicBlock());
    llvm::BranchInst::Create(thenBlock, elseBlock, condition, b);

    llvm::BranchInst::Create(footBlock, irgen.GetBasicBlock());

    irgen.SetBasicBlock(elseBlock);
    llvm::Value* fVal = falseExpr->Emit();
    new llvm::StoreInst(fVal, val, false, irgen.GetBasicBlock());

    llvm::BranchInst::Create(footBlock, irgen.GetBasicBlock());
    irgen.SetBasicBlock(footBlock);
    return new llvm::LoadInst(val, "", footBlock);
}

static llvm::Value* typeConvert(llvm::Value* val, llvm::Value* target) {
    if(val->getType() == target->getType()) return val;
    llvm::BasicBlock* b = irgen.GetBasicBlock();
    llvm::Value* ret = new llvm::AllocaInst(target->getType(), "", b);
    ret = new llvm::LoadInst(ret, "", b);

    for(int i = 0; i < target->getType()->getVectorNumElements(); i++) {
        ret = llvm::InsertElementInst::Create(ret, val, llvm::ConstantInt::get(irgen.GetIntType(), i), "", b);
    }
    return ret;
}

llvm::Value* CompoundExpr::Emit() {
    llvm::BasicBlock* b = irgen.GetBasicBlock();
    llvm::Function * func = irgen.GetFunction();
    llvm::LLVMContext* context = irgen.GetContext();
    //weird bc of short circuting
    if(op->getToken() == "&&" || op->getToken() == "||") {
        llvm::Value* val = new llvm::AllocaInst(llvm::Type::getInt1Ty(*context), "", b);
        llvm::BasicBlock * otherBlock = llvm::BasicBlock::Create(*context, "otherBlock", func);
        llvm::BasicBlock * falseBlock = llvm::BasicBlock::Create(*context, "falseBlock", func);
        llvm::BasicBlock * trueBlock = llvm::BasicBlock::Create(*context, "trueBlock", func);
        llvm::BasicBlock * footBlock = llvm::BasicBlock::Create(*context, "footBlock", func);
        llvm::Value* l = left->Emit();
        b = irgen.GetBasicBlock();
        if(op->getToken() == "&&") llvm::BranchInst::Create(otherBlock, falseBlock, l, b);
        if(op->getToken() == "||") llvm::BranchInst::Create(trueBlock, otherBlock, l, b);
        irgen.SetBasicBlock(otherBlock);
        llvm::Value* r = right->Emit();
        b = irgen.GetBasicBlock();
        llvm::BranchInst::Create(trueBlock, falseBlock, r, b);
        new llvm::StoreInst(llvm::ConstantInt::get(llvm::Type::getInt1Ty(*context), true), val, false, trueBlock);
        llvm::BranchInst::Create(footBlock, trueBlock);
        new llvm::StoreInst(llvm::ConstantInt::get(llvm::Type::getInt1Ty(*context), false), val, false, falseBlock);
        llvm::BranchInst::Create(footBlock, falseBlock);
        irgen.SetBasicBlock(footBlock);
        return new llvm::LoadInst(val, "", footBlock);
    }
    llvm::Value* l = left ? left->Emit() : NULL;
    llvm::Value* r = right ? right->Emit() : NULL;
    b = irgen.GetBasicBlock();
    //austins additional value
    llvm::Value *val;
    if(!l || !r) {
        if(op->getToken() == "-" && (l ? l->getType()->isFPOrFPVectorTy() : r->getType()->isFPOrFPVectorTy())) {
            return llvm::BinaryOperator::Create(llvm::Instruction::FSub, typeConvert(llvm::ConstantFP::get(irgen.GetFloatType(), 0.0), (l?l:r)), l ? l : r, "", b);
        }
        else if(op->getToken() == "-") {
            return llvm::BinaryOperator::Create(llvm::Instruction::Sub, typeConvert(llvm::ConstantInt::get(irgen.GetIntType(), 0), (l?l:r)), l ? l : r, "", b);
        }
        else if(op->getToken() == "+") return l ? l : r;
        VarExpr* ve = dynamic_cast<VarExpr*>(l ? left : right);
        llvm::Value* ret;
        llvm::Value* var = l ? l : r;

        if(var->getType()->isFPOrFPVectorTy()) val = llvm::ConstantFP::get(irgen.GetFloatType(), 1.0);
        else val = llvm::ConstantInt::get(irgen.GetIntType(), 1);
        val = typeConvert(val, var);

        if(op->getToken() == "++") {
            ret = llvm::BinaryOperator::CreateAdd(var, val, "", b);
        }
        else if(op->getToken() == "--") {
            ret = llvm::BinaryOperator::CreateSub(var, val, "", b);
        }
        new llvm::StoreInst(ret, findSymbol(ve->GetIdentifier()->GetName()).second, false, b);

        if(!l) return ret;
        else return var;
    }
    if(l->getType()->isVectorTy() && !r->getType()->isVectorTy()) r = typeConvert(r, l);
    if(!l->getType()->isVectorTy() && r->getType()->isVectorTy()) l = typeConvert(l, r);
    if(l->getType()->isFPOrFPVectorTy()) {
        if(op->getToken() == "+")
            return llvm::BinaryOperator::Create(llvm::Instruction::FAdd, l, r, "", b);
        if(op->getToken() == "-")
            return llvm::BinaryOperator::Create(llvm::Instruction::FSub, l, r, "", b);
        if(op->getToken() == "*")
            return llvm::BinaryOperator::Create(llvm::Instruction::FMul, l, r, "", b);
        if(op->getToken() == "/")
            return llvm::BinaryOperator::Create(llvm::Instruction::FDiv, l, r, "", b);
        if(op->getToken() == "==")
            return llvm::CmpInst::Create(llvm::CmpInst::FCmp, llvm::CmpInst::FCMP_OEQ, l, r, "", b);
        if(op->getToken() == "!=")
            return llvm::CmpInst::Create(llvm::CmpInst::FCmp, llvm::CmpInst::FCMP_ONE, l, r, "", b);
        if(op->getToken() == "<=")
            return llvm::CmpInst::Create(llvm::CmpInst::FCmp, llvm::CmpInst::FCMP_OLE, l, r, "", b);
        if(op->getToken() == ">=")
            return llvm::CmpInst::Create(llvm::CmpInst::FCmp, llvm::CmpInst::FCMP_OGE, l, r, "", b);
        if(op->getToken() == "<")
            return llvm::CmpInst::Create(llvm::CmpInst::FCmp, llvm::CmpInst::FCMP_OLT, l, r, "", b);
        if(op->getToken() == ">")
            return llvm::CmpInst::Create(llvm::CmpInst::FCmp, llvm::CmpInst::FCMP_OGT, l, r, "", b);
    }
    if(l->getType()->isIntOrIntVectorTy()) {
        if(op->getToken() == "+")
            return llvm::BinaryOperator::Create(llvm::Instruction::Add, l, r, "", b);
        if(op->getToken() == "-")
            return llvm::BinaryOperator::Create(llvm::Instruction::Sub, l, r, "", b);
        if(op->getToken() == "*")
            return llvm::BinaryOperator::Create(llvm::Instruction::Mul, l, r, "", b);
        if(op->getToken() == "/")
            return llvm::BinaryOperator::Create(llvm::Instruction::SDiv, l, r, "", b);
        if(op->getToken() == "==")
            return llvm::CmpInst::Create(llvm::CmpInst::ICmp, llvm::CmpInst::ICMP_EQ, l, r, "", b);
        if(op->getToken() == "!=")
            return llvm::CmpInst::Create(llvm::CmpInst::ICmp, llvm::CmpInst::ICMP_NE, l, r, "", b);
        if(op->getToken() == "<=")
            return llvm::CmpInst::Create(llvm::CmpInst::ICmp, llvm::CmpInst::ICMP_SLE, l, r, "", b);
        if(op->getToken() == ">=")
            return llvm::CmpInst::Create(llvm::CmpInst::ICmp, llvm::CmpInst::ICMP_SGE, l, r, "", b);
        if(op->getToken() == "<")
            return llvm::CmpInst::Create(llvm::CmpInst::ICmp, llvm::CmpInst::ICMP_SLT, l, r, "", b);
        if(op->getToken() == ">")
            return llvm::CmpInst::Create(llvm::CmpInst::ICmp, llvm::CmpInst::ICMP_SGT, l, r, "", b);
    }
    return NULL;
}

//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------

llvm::Value* AssignExpr::Emit(){
    llvm::Value * leftval= left->Emit();
    llvm::Value * rightval=right->Emit();
    llvm::LoadInst* leftLookup= llvm::cast<llvm::LoadInst>(leftval);
    llvm::BasicBlock* basicBlock= irgen.GetBasicBlock();
    llvm::Value *targetL = leftLookup->getPointerOperand();
    llvm::Value * retVal = NULL;
    llvm::Value *binOp = NULL;

    if(dynamic_cast<FieldAccess*>(left)!=NULL){
        //std::cerr<< "field access"<<"\n";
        return (dynamic_cast<FieldAccess*>(left)->FieldStore(rightval,op->getToken()[0]));
    }
    if(op->IsOp("=")){
        new llvm::StoreInst(rightval,targetL,basicBlock);
        retVal=rightval;
    }
    else if(op->IsOp("+=")){
        if( leftval->getType()== convertType(Type::intType)){
            binOp=llvm::BinaryOperator::CreateAdd(leftval,rightval,"BinaryInt AddEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
        if(leftval->getType()==convertType(Type::floatType)){
            binOp=llvm::BinaryOperator::CreateFAdd(leftval,rightval,"BinaryFloat AddEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
        if(leftval->getType()==convertType(Type::vec2Type)){
            binOp=llvm::BinaryOperator::CreateFAdd(leftval,rightval,"Vec2 AddEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
        if(leftval->getType()==convertType(Type::vec3Type)){
            binOp=llvm::BinaryOperator::CreateFAdd(leftval,rightval,"Vec3 AddEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
        if(leftval->getType()==convertType(Type::vec4Type)){
            binOp=llvm::BinaryOperator::CreateFAdd(leftval,rightval,"Vec4 AddEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
    }
    else if(op->IsOp("-=")){
        if(leftval->getType()==convertType(Type::intType)){
            binOp=llvm::BinaryOperator::CreateSub(leftval,rightval,"BinaryInt SubEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
        if(leftval->getType()==convertType(Type::floatType)){
            binOp=llvm::BinaryOperator::CreateFSub(leftval,rightval,"BinaryFloat SubEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
        if(leftval->getType()==convertType(Type::vec2Type)){
            binOp=llvm::BinaryOperator::CreateFSub(leftval,rightval,"Vec2 SubEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
        if(leftval->getType()==convertType(Type::vec3Type)){
            binOp=llvm::BinaryOperator::CreateFSub(leftval,rightval,"Vec3 SubEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
        if(leftval->getType()==convertType(Type::vec4Type)){
            binOp=llvm::BinaryOperator::CreateSub(leftval,rightval,"Vec4 SubEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
    }
    else if(op->IsOp("*=")){
        if(leftval->getType()==convertType(Type::intType)){
            binOp=llvm::BinaryOperator::CreateMul(leftval,rightval,"BinaryInt MulEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
        if(leftval->getType()==convertType(Type::floatType)){
            binOp=llvm::BinaryOperator::CreateFMul(leftval,rightval,"BinaryFloat MulEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
        if(leftval->getType()==convertType(Type::vec2Type)){
            binOp=llvm::BinaryOperator::CreateFMul(leftval,rightval,"Vec2 MulEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
        if(leftval->getType()==convertType(Type::vec3Type)){
            binOp=llvm::BinaryOperator::CreateFMul(leftval,rightval,"Vec3 MulEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
        if(leftval->getType()==convertType(Type::vec4Type)){
            binOp=llvm::BinaryOperator::CreateFMul(leftval,rightval,"Vec4 MulEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
    }
    else if(op->IsOp("/=")){
        if(leftval->getType()==convertType(Type::intType)){
            binOp=llvm::BinaryOperator::CreateSDiv(leftval,rightval,"BinaryInt DivEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
        if(leftval->getType()==convertType(Type::floatType)){
            binOp=llvm::BinaryOperator::CreateFDiv(leftval,rightval,"BinaryFloat DivEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
        if(leftval->getType()==convertType(Type::vec2Type)){
            binOp=llvm::BinaryOperator::CreateFDiv(leftval,rightval,"Vec2 DivEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
        if(leftval->getType()==convertType(Type::vec3Type)){
            binOp=llvm::BinaryOperator::CreateFDiv(leftval,rightval,"Vec3 DivEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
        if(leftval->getType()==convertType(Type::vec4Type)){
            binOp=llvm::BinaryOperator::CreateFDiv(leftval,rightval,"Vec4 DivEqual", basicBlock);
            new llvm::StoreInst(binOp,targetL,basicBlock);
        }
    }
    return rightval;
}


llvm::Value * FieldAccess::FieldStore(llvm::Value* val, char oper){
    llvm::Value* baseVal = getBase()->Emit();
    string fieldName = getField()->GetName();
    llvm::Constant* constVal=NULL;
    llvm::Value* exLeftElmt= NULL;
    llvm::Value* exRightElmt= NULL;
    llvm::Value* operVal = NULL;
    if(fieldName.length()==1){
        char charVal=fieldName.at(0);
        exLeftElmt= llvm::ExtractElementInst::Create(baseVal,FindChar(charVal),"ExtractedVal",irgen.GetBasicBlock());
        if(oper=='+'){
            operVal=llvm::BinaryOperator::CreateFAdd(exLeftElmt,val,"Field FAdd",irgen.GetBasicBlock());
        }
        else if(oper=='-'){
            operVal=llvm::BinaryOperator::CreateFSub(exLeftElmt,val,"Field FSub",irgen.GetBasicBlock());
        }
        else if(oper=='*'){
            operVal=llvm::BinaryOperator::CreateFMul(exLeftElmt,val,"Field FMul",irgen.GetBasicBlock());
        }
        else if(oper=='/'){
            operVal=llvm::BinaryOperator::CreateFDiv(exLeftElmt,val,"Field FDiv",irgen.GetBasicBlock());
        }
        else{
            operVal=val;
        }
        baseVal= llvm::InsertElementInst::Create(baseVal,operVal,FindChar(charVal),"Base",irgen.GetBasicBlock());
    }
    else{
        for(int i=0; i<fieldName.length();i++){
            char charVal= fieldName.at(i);
            constVal = llvm::ConstantInt::get(irgen.GetIntType(),i, false);
            exRightElmt=llvm::ExtractElementInst::Create(val,constVal,"ExtractedRightVal",irgen.GetBasicBlock());
            exLeftElmt=llvm::ExtractElementInst::Create(baseVal,constVal,"ExtractedVal",irgen.GetBasicBlock());
            if(oper=='+'){
                operVal=llvm::BinaryOperator::CreateFAdd(exLeftElmt,val,"Field FAdd",irgen.GetBasicBlock());
            }
            else if(oper=='-'){
                operVal=llvm::BinaryOperator::CreateFSub(exLeftElmt,val,"Field FSub",irgen.GetBasicBlock());
            }
            else if(oper=='*'){
                operVal=llvm::BinaryOperator::CreateFMul(exLeftElmt,val,"Field FMul",irgen.GetBasicBlock());
            }
            else if(oper=='/'){
                operVal=llvm::BinaryOperator::CreateFDiv(exLeftElmt,val,"Field FDiv",irgen.GetBasicBlock());
            }
            else{
                operVal=exRightElmt;
            }
            baseVal=llvm::InsertElementInst::Create(baseVal,operVal,FindChar(charVal),"Base",irgen.GetBasicBlock());
        }
    }
    //dynamic_cast<VarExpr*>(base)->StoreVal(baseVal);
    return val;
}

llvm::Value* FieldAccess::FindChar(char c){
    llvm::Value* CharValue=NULL;
    if(c=='x'){
        CharValue=llvm::ConstantInt::get(irgen.GetIntType(),0);
    }
    else if(c=='y'){
        CharValue=llvm::ConstantInt::get(irgen.GetIntType(),1);
    }
    else if(c=='z'){
        CharValue=llvm::ConstantInt::get(irgen.GetIntType(),2);
    }
    else if(c=='w'){
        CharValue=llvm::ConstantInt::get(irgen.GetIntType(),3);
    }
    return CharValue;
}

llvm::Value * AssignExpr::convert(llvm::Value * vec, llvm::Value * val, int status ){
    llvm::Type * intTyp = irgen.GetIntType();
    llvm::Type * fltTyp = irgen.GetFloatType();
    llvm::Value *idVal;
    llvm::Value * element;
    llvm::Value * retVal = vec;

    //get vector size
    int numElements=-1;
    if(vec->getType()==irgen.GetVec2Type()){
        numElements=2;
    }
    else if(vec->getType()==irgen.GetVec3Type()){
        numElements=3;
    }
    else if(vec->getType()==irgen.GetVec4Type()){
        numElements=4;
    }

    for (int i = 0; i < numElements; i++){
        idVal = llvm::ConstantInt::get(intTyp, i);
        element = llvm::ExtractElementInst::Create (vec, idVal, "", irgen.GetBasicBlock());
        if (status == 0){
            element = llvm::BinaryOperator::CreateFAdd(element, val, "", irgen.GetBasicBlock());
        }
        else if (status == 1){
            element = llvm::BinaryOperator::CreateFSub(element, val, "", irgen.GetBasicBlock());
        }
        else if (status == 2){
            element = llvm::BinaryOperator::CreateFMul(element, val, "", irgen.GetBasicBlock());
        }
        else{
            element = llvm::BinaryOperator::CreateFDiv(element, val, "", irgen.GetBasicBlock());
        }
        retVal = llvm::InsertElementInst::Create (retVal, element, idVal, "", irgen.GetBasicBlock());
    }
    return retVal;
}

llvm::Value* VarExpr::AddressStore(llvm::Value *store){
    char * varname = GetIdentifier()->GetName();

    pair<Decl *, llvm::Value *> symbol = findSymbol(string(varname));
    //value of symbol
    llvm::StoreInst* storeInst = new llvm::StoreInst(store, symbol.second, false, irgen.GetBasicBlock());
    return storeInst;
}

llvm::Value * AssignExpr::AssignField(){
    FieldAccess * f = (FieldAccess *) left;
    llvm::Value * rightval = right->Emit();
    string operationStr(op->getToken());
    bool isFloat = rightval->getType()->isFloatingPointTy();
    llvm::Type * intTyp = irgen.GetIntType();
    char * ptr = f->getField()->GetName();;

    std::vector<llvm::Constant *> maskI;
    for(; *ptr; ptr++){
        if(*ptr =='x'){
            maskI.push_back(llvm::ConstantInt::get(intTyp, 0));
        }
        else if(*ptr='y'){
            maskI.push_back(llvm::ConstantInt::get(intTyp, 1));
        }
        else if(*ptr='z'){
            maskI.push_back(llvm::ConstantInt::get(intTyp, 2));
        }
        else if(*ptr='w'){
            maskI.push_back(llvm::ConstantInt::get(intTyp, 3));
        }
    }

    llvm::Value * element;
    llvm::Value * rightop = rightval;
    llvm::Value * baseF = f->getBase()->Emit();
    llvm::Value * storeInst = f->getBase()->Emit();


    for (int i = 0; i < maskI.size(); i++){
        element = llvm::ExtractElementInst::Create (baseF, maskI[i], "", irgen.GetBasicBlock());
        if (rightval->getType()->isVectorTy()){
            rightop = llvm::ExtractElementInst::Create(rightval, llvm::ConstantInt::get(intTyp, i), "", irgen.GetBasicBlock());
        }
        if (!operationStr.compare("=")){
            storeInst = llvm::InsertElementInst::Create (storeInst, rightop, maskI[i], "", irgen.GetBasicBlock());
        }
        if (!operationStr.compare("+=")){
            element = llvm::BinaryOperator::CreateFAdd(element, rightop, "", irgen.GetBasicBlock());
            storeInst = llvm::InsertElementInst::Create (storeInst, element, maskI[i], "", irgen.GetBasicBlock());
        }
        if (!operationStr.compare("-=")){
            element = llvm::BinaryOperator::CreateFSub(element, rightop, "", irgen.GetBasicBlock());
            storeInst = llvm::InsertElementInst::Create (storeInst, element, maskI[i], "", irgen.GetBasicBlock());
        }
        if (!operationStr.compare("*=")){
            element = llvm::BinaryOperator::CreateFMul(element, rightop, "", irgen.GetBasicBlock());
            storeInst = llvm::InsertElementInst::Create (storeInst, element, maskI[i], "", irgen.GetBasicBlock());
        }
        if (!operationStr.compare("/=")){
            element = llvm::BinaryOperator::CreateFDiv(element, rightop, "", irgen.GetBasicBlock());
            storeInst = llvm::InsertElementInst::Create (storeInst, element, maskI[i], "", irgen.GetBasicBlock());
        }
    }
    VarExpr * ve = dynamic_cast<VarExpr *>(f->getBase());
    ve->AddressStore(storeInst);
    if (isFloat){
        return llvm::ExtractElementInst::Create (storeInst, maskI.back(), "", irgen.GetBasicBlock());
    }
    return storeInst;
}


llvm::Value * FieldAccess::Emit(){
    char * mask = this->getField()->GetName();
    char * ptr = mask;

    llvm::Type * intTyp = irgen.GetIntType();
    std::vector<llvm::Constant *> maskI;

    int count = 0;
    int valNum = -1;
    for(; *ptr; ptr++){
        if(*ptr =='x'){
            maskI.push_back(llvm::ConstantInt::get(intTyp, 0));
            valNum=0;
        }
        else if(*ptr=='y'){
            maskI.push_back(llvm::ConstantInt::get(intTyp, 1));
            valNum=1;
        }
        else if(*ptr=='z'){
            maskI.push_back(llvm::ConstantInt::get(intTyp, 2));
            valNum=2;
        }
        else if(*ptr=='w'){
            maskI.push_back(llvm::ConstantInt::get(intTyp, 3));
            valNum=3;
        }
        count++;
    }
    llvm::Value * elemt;
    llvm::Value * baseF = base->Emit();
    if (count == 1){
        elemt = llvm::ExtractElementInst::Create (baseF, maskI.back(), "", irgen.GetBasicBlock());
        return elemt;
    }
    llvm::ConstantVector *masky = (llvm::ConstantVector *) llvm::ConstantVector::get(maskI);
    llvm::Value * val = new llvm::ShuffleVectorInst(baseF,llvm::UndefValue::get(baseF->getType()), masky,"",irgen.GetBasicBlock());
    return val;
}

llvm::Value * ArrayAccess::Emit(){
    Expr * base= GetBase();
    Expr * sub= GetSubScript();
    //std::cerr<< "arrayAccess"<<"\n";

    if(dynamic_cast< VarExpr* >(base)->GetIdentifier()->GetName()){
        llvm::Value * val= findSymbol(dynamic_cast< VarExpr* >(base)->GetIdentifier()->GetName()).second;
        llvm::Value * subscrVal = sub->Emit();
        std::vector <llvm::Value*> vect;
        vect.push_back(llvm::ConstantInt::get(irgen.GetIntType(),0));
        vect.push_back((llvm::Value *const &) subscrVal);
        llvm::ArrayRef<llvm::Value *> valList(vect);
        llvm::Value *elmt= llvm::GetElementPtrInst::Create(val,valList,"ARRAYACCESS",irgen.GetBasicBlock());

        return new llvm::LoadInst(elmt,"ARRAYACCESS",irgen.GetBasicBlock());
    }
    else{
        //std::cerr<< "fucking fuck fuck fuck"<<"\n";
        return NULL;
    }
}


