/* File: ast_stmt.cc
 * -----------------
 * Implementation of statement node classes.
 */
#include "ast_stmt.h"
#include "ast_type.h"
#include "ast_decl.h"
#include "ast_expr.h"
#include "symtable.h"

#include "irgen.h"
#include "llvm/Bitcode/ReaderWriter.h"
#include "llvm/Support/raw_ostream.h"                                                   


Program::Program(List<Decl*> *d) {
    Assert(d != NULL);
    (decls=d)->SetParentAll(this);
}

void Program::PrintChildren(int indentLevel) {
    decls->PrintAll(indentLevel+1);
    printf("\n");
}
//pls work
llvm::Value* Program::Emit() {
    llvm::Module *module = irgen.GetOrCreateModule("program");
    pushScope();
    for (int i = 0; i < decls->NumElements(); i++){
        decls->Nth(i)->Emit();
    }
    popScope();

    module->dump();
    llvm::WriteBitcodeToFile(module, llvm::outs());
    return NULL;
}

StmtBlock::StmtBlock(List<VarDecl*> *d, List<Stmt*> *s) {
    Assert(d != NULL && s != NULL);
    (decls=d)->SetParentAll(this);
    (stmts=s)->SetParentAll(this);
}

void StmtBlock::PrintChildren(int indentLevel) {
    decls->PrintAll(indentLevel+1);
    stmts->PrintAll(indentLevel+1);
}

DeclStmt::DeclStmt(Decl *d) {
    Assert(d != NULL);
    (decl=d)->SetParent(this);
}

void DeclStmt::PrintChildren(int indentLevel) {
    decl->Print(indentLevel+1);
}

ConditionalStmt::ConditionalStmt(Expr *t, Stmt *b) { 
    Assert(t != NULL && b != NULL);
    (test=t)->SetParent(this); 
    (body=b)->SetParent(this);
}

ForStmt::ForStmt(Expr *i, Expr *t, Expr *s, Stmt *b): LoopStmt(t, b) { 
    Assert(i != NULL && t != NULL && b != NULL);
    (init=i)->SetParent(this);
    step = s;
    if ( s )
      (step=s)->SetParent(this);
}

void ForStmt::PrintChildren(int indentLevel) {
    init->Print(indentLevel+1, "(init) ");
    test->Print(indentLevel+1, "(test) ");
    if ( step )
      step->Print(indentLevel+1, "(step) ");
    body->Print(indentLevel+1, "(body) ");
}

void WhileStmt::PrintChildren(int indentLevel) {
    test->Print(indentLevel+1, "(test) ");
    body->Print(indentLevel+1, "(body) ");
}

IfStmt::IfStmt(Expr *t, Stmt *tb, Stmt *eb): ConditionalStmt(t, tb) { 
    Assert(t != NULL && tb != NULL); // else can be NULL
    elseBody = eb;
    if (elseBody) elseBody->SetParent(this);
}

void IfStmt::PrintChildren(int indentLevel) {
    if (test) test->Print(indentLevel+1, "(test) ");
    if (body) body->Print(indentLevel+1, "(then) ");
    if (elseBody) elseBody->Print(indentLevel+1, "(else) ");
}


ReturnStmt::ReturnStmt(yyltype loc, Expr *e) : Stmt(loc) { 
    expr = e;
    if (e != NULL) expr->SetParent(this);
}

void ReturnStmt::PrintChildren(int indentLevel) {
    if ( expr ) 
      expr->Print(indentLevel+1);
}

SwitchLabel::SwitchLabel(Expr *l, Stmt *s) {
    Assert(l != NULL && s != NULL);
    (label=l)->SetParent(this);
    (stmt=s)->SetParent(this);
}

SwitchLabel::SwitchLabel(Stmt *s) {
    Assert(s != NULL);
    label = NULL;
    (stmt=s)->SetParent(this);
}

void SwitchLabel::PrintChildren(int indentLevel) {
    if (label) label->Print(indentLevel+1);
    if (stmt)  stmt->Print(indentLevel+1);
}

SwitchStmt::SwitchStmt(Expr *e, List<Stmt *> *c, Default *d) {
    Assert(e != NULL && c != NULL && c->NumElements() != 0 );
    (expr=e)->SetParent(this);
    (cases=c)->SetParentAll(this);
    def = d;
    if (def) def->SetParent(this);
}

void SwitchStmt::PrintChildren(int indentLevel) {
    if (expr) expr->Print(indentLevel+1);
    if (cases) cases->PrintAll(indentLevel+1);
    if (def) def->Print(indentLevel+1);
}
//-----------------------------------------------------------------------
//rest of the emits
//-----------------------------------------------------------------------
llvm::Value * StmtBlock::Emit(){
    pushScope();
    for (int i = 0; i < decls->NumElements(); i++){
        decls->Nth(i)->Emit();
    }
    for (int i = 0; i < stmts->NumElements(); i++){
        stmts->Nth(i)->Emit();
        if(irgen.GetBasicBlock()->getTerminator()) return NULL;
    }

    popScope();
    return NULL;
}

llvm::Value * DeclStmt::Emit(){
    llvm::Value * val;
    if (VarDecl * vd = dynamic_cast<VarDecl*>(this->decl)){
        val = vd->Emit();
    }
    else if (FnDecl * fd = dynamic_cast<FnDecl*>(this->decl)){
        val = fd->Emit();
    }
    else{
        val = NULL;
    }
    return val;
}

llvm::Value * ConditionalStmt::Emit(){
    return NULL;
}

llvm::Value * ForStmt::Emit()
{
    pushScope();
    llvm::LLVMContext * context = irgen.GetContext();
    llvm::Function * f = irgen.GetFunction();
    llvm::BasicBlock * headBlock = llvm::BasicBlock::Create(*context, "headBlock", f);
    llvm::BasicBlock * footBlock = llvm::BasicBlock::Create(*context, "footBlock", f);
    llvm::BasicBlock * stepBlock = llvm::BasicBlock::Create(*context, "stepBlock", f);
    llvm::BasicBlock * bodyBlock = llvm::BasicBlock::Create(*context, "bodyBlock", f);
    llvm::Value * cond;

    cout<<endl;
    // init and branch to head
    init->Emit();

    llvm::BranchInst::Create(headBlock, irgen.GetBasicBlock()); // given in the slides
    llvm::BasicBlock * oldHeader = irgen.GetHeaderBlock();
    llvm::BasicBlock * oldFooter = irgen.GetFooterBlock();
    irgen.SetHeaderBlock(stepBlock);
    irgen.SetFooterBlock(footBlock);

    // head
    irgen.SetBasicBlock(headBlock);
    cond = test->Emit();
    if(!irgen.GetBasicBlock()->getTerminator())
    {
        llvm::BranchInst::Create(bodyBlock, footBlock, cond, irgen.GetBasicBlock()); // given in the slides
    }
    // body
    irgen.SetBasicBlock(bodyBlock);
    body->Emit();
    if(!irgen.GetBasicBlock()->getTerminator())
    {
        llvm::BranchInst::Create(stepBlock, irgen.GetBasicBlock()); // given in the slides
    }
    // step
    irgen.SetBasicBlock(stepBlock);
    step->Emit();
    if(!irgen.GetBasicBlock()->getTerminator())
    {
        llvm::BranchInst::Create(headBlock, irgen.GetBasicBlock()); // given in the slides
    }
    irgen.SetBasicBlock(footBlock);
    irgen.SetHeaderBlock(oldHeader);
    irgen.SetFooterBlock(oldFooter);
    popScope();
    return NULL;
}

//while  statement
llvm::Value * WhileStmt::Emit(){
    pushScope();
    llvm::LLVMContext * context = irgen.GetContext();
    llvm::Function * f = irgen.GetFunction();
    llvm::Value * cond;

    llvm::BasicBlock * headBlock = llvm::BasicBlock::Create(*context, "headBlock", f);
    llvm::BasicBlock * footBlock = llvm::BasicBlock::Create(*context, "footBlock", f);
    llvm::BasicBlock * bodyBlock = llvm::BasicBlock::Create(*context, "bodyBlock", f);
    llvm::BranchInst::Create(headBlock, irgen.GetBasicBlock());
    llvm::BasicBlock * oldHeader = irgen.GetHeaderBlock();
    llvm::BasicBlock * oldFooter = irgen.GetFooterBlock();
    irgen.SetHeaderBlock(headBlock);
    irgen.SetFooterBlock(footBlock);

    irgen.SetBasicBlock(headBlock);
    cond = test->Emit();
    if(!irgen.GetBasicBlock()->getTerminator())
    {
        llvm::BranchInst::Create(bodyBlock, footBlock, cond, irgen.GetBasicBlock()); // given in the slides
    }

    irgen.SetBasicBlock(bodyBlock);
    body->Emit();
    if(!irgen.GetBasicBlock()->getTerminator())
    {
        llvm::BranchInst::Create(headBlock, irgen.GetBasicBlock()); // given in the slides
    }

    irgen.SetBasicBlock(footBlock);
    irgen.SetHeaderBlock(oldHeader);
    irgen.SetFooterBlock(oldFooter);
    popScope();
    return NULL;
}


//if statement
llvm::Value * IfStmt::Emit(){

    llvm::LLVMContext * context = irgen.GetContext();
    llvm::Function * func = irgen.GetFunction();
    llvm::BasicBlock * elseBlock = NULL;
    llvm::BasicBlock * thenBlock = llvm::BasicBlock::Create(*context, "thenBlock", func);
    llvm::BasicBlock * footBlock = llvm::BasicBlock::Create(*context, "footBlock", func);
    llvm::Value * val;
    llvm::Value * condition = test->Emit();

    if(elseBody)
    {
        elseBlock =  llvm::BasicBlock::Create(*context, "elseBlock", func);
    }

    val = llvm::BranchInst::Create(thenBlock, elseBody ? elseBlock : footBlock, condition, irgen.GetBasicBlock());
    pushScope();
    irgen.SetBasicBlock(thenBlock);
    body->Emit();

    if(!irgen.GetBasicBlock()->getTerminator())
    {
        val = llvm::BranchInst::Create(footBlock, irgen.GetBasicBlock());
    }

    popScope();

    if(elseBody)
    {
        pushScope();
        irgen.SetBasicBlock(elseBlock);
        elseBody->Emit();

        if(!irgen.GetBasicBlock()->getTerminator())
        {
            val = llvm::BranchInst::Create(footBlock, irgen.GetBasicBlock());
        }
        popScope();
    }
    irgen.SetBasicBlock(footBlock);
    return val;
}

llvm::Value * BreakStmt::Emit(){
    //goes to footer
    llvm::Value * val;
    llvm::LLVMContext * context = irgen.GetContext();
    llvm::Function * func = irgen.GetFunction();
    val=llvm::BranchInst::Create (irgen.GetFooterBlock(), irgen.GetBasicBlock());
    return val;
}

llvm::Value * ContinueStmt::Emit(){
    llvm::Value * val;
    llvm::LLVMContext * context = irgen.GetContext();
    llvm::Function * func = irgen.GetFunction();
    val=llvm::BranchInst::Create (irgen.GetHeaderBlock(), irgen.GetBasicBlock());
    return val;
}

//Not sure
llvm::Value * ReturnStmt::Emit(){
    llvm::Value * val;
    llvm::LLVMContext * context = irgen.GetContext();
    if (expr){
        llvm::Value * retVal = expr->Emit();
        retVal = llvm::ReturnInst::Create(*context, retVal, irgen.GetBasicBlock());
        return retVal;
    }
    val = llvm::ReturnInst::Create(*context, irgen.GetBasicBlock());
    return val;
}
//stack for both

llvm::SwitchInst * curSwitch = NULL;
llvm::Value * SwitchStmt::Emit(){
    llvm::SwitchInst * val;
    llvm::LLVMContext * context = irgen.GetContext();
    llvm::Function * func = irgen.GetFunction();
    vector<llvm::BasicBlock *> blockArr;
    int defaultIndex = -1;
    //find all cases and create a basic block for all of them
    for (int i = 0; i < cases->NumElements(); i++){
        blockArr.push_back(llvm::BasicBlock::Create(*context, "switchBlock", func));
        if(dynamic_cast<Default*>(cases->Nth(i))) defaultIndex = i;
    }
    //Emit of expression
    llvm::Value * exprVal=expr->Emit();
    llvm::BasicBlock* footer = llvm::BasicBlock::Create(*context, "switchFooter", func);
    llvm::BasicBlock * oldFooter = irgen.GetFooterBlock();
    irgen.SetFooterBlock(footer);

    //create switch instruction pass in default
    llvm::BasicBlock* defaultBlock = defaultIndex >= 0 ? blockArr[defaultIndex] : footer;
    val=llvm::SwitchInst::Create(exprVal, defaultBlock, cases->NumElements(), irgen.GetBasicBlock());
    llvm::SwitchInst* oldSwitch = curSwitch;
    curSwitch = val;
    pushScope();

    //for each case 'addCase to switch stmt', EMIT for stmt in case, create terminator instruction
    for(int i=0; i<cases->NumElements();i++){
        //add addCase to switch stmt
        //addcase to val method
        irgen.SetBasicBlock(blockArr[i]);
        cases->Nth(i)->Emit();
        if(!blockArr[i]->getTerminator()) {
            if(i+1 < cases->NumElements()) llvm::BranchInst::Create(blockArr[i+1], blockArr[i]);
            else llvm::BranchInst::Create(footer, blockArr[i]);
        }
        if(!irgen.GetBasicBlock()->getTerminator()) {
            if(i+1 < cases->NumElements()) llvm::BranchInst::Create(blockArr[i+1], irgen.GetBasicBlock());
            else llvm::BranchInst::Create(footer, irgen.GetBasicBlock());
        }
    }
    irgen.SetBasicBlock(footer);
    irgen.SetFooterBlock(oldFooter);
    curSwitch = oldSwitch;
    popScope();
    return val;
}

llvm::Value * SwitchLabel::Emit(){
    return NULL;
}

llvm::Value * Case::Emit(){
    //stmt and label emit on stmt //return on label
    curSwitch->addCase (dynamic_cast<llvm::ConstantInt*>(label->Emit()), irgen.GetBasicBlock());
    return stmt->Emit();
}

llvm::Value * Default::Emit(){
    //stmt emit return null
    stmt->Emit();
    return NULL;
}

